"""Forms for the file_validator app."""

import csv
from collections.abc import Iterable, Sequence
from typing import ClassVar

from django import forms
from django.core.files.uploadedfile import UploadedFile


class CSVUploadForm(forms.Form):
    """Form for uploading CSV files only."""

    file = forms.FileField(label="Select a CSV file")

    required_columns: ClassVar[list] = [
        "AccountNumber",
        "CBAccountNumber",
        "DaysDiscountValid",
        "DiscountValue",
        "DiscountPercentage",
        "DueDate",
        "GoodsValueInAccountCurrency",
        "PurControlValueInBaseCurrency",
        "DocumentToBaseCurrencyRate",
        "DocumentToAccountCurrencyRate",
        "PostedDate",
        "QueryCode",
        "TransactionReference",
        "SecondReference",
        "Source",
        "SYSTraderTranType",
        "TransactionDate",
        "UniqueReferenceNumber",
        "UserNumber",
        "TaxValue",
        "SYSTraderGenerationReasonType",
        "GoodsValueInBaseCurrency",
    ]

    repeating_columns: ClassVar[dict] = {
        "NominalAnalysis": [
            "NominalAnalysisTransactionValue",
            "NominalAnalysisNominalAccountNumber",
            "NominalAnalysisNominalCostCentre",
            "NominalAnalysisNominalDepartment",
            "NominalAnalysisNominalAnalysisNarrative",
            "NominalAnalysisTransactionAnalysisCode",
        ],
        "TaxAnalysis": [
            "TaxAnalysisTaxRate",
            "TaxAnalysisGoodsValueBeforeDiscount",
            "TaxAnalysisDiscountValue",
            "TaxAnalysisDiscountPercentage",
            "TaxAnalysisTaxOnGoodsValue",
        ],
    }

    def clean_file(self) -> UploadedFile:
        """Validate the uploaded file."""
        file = self.cleaned_data["file"]

        # Step 1: Validate file type
        self._validate_file_type(file)

        # Step 2: Parse file and validate headers
        csv_file = file.read().decode("utf-8").splitlines()
        reader = csv.DictReader(csv_file, delimiter=";")
        fieldnames = reader.fieldnames if reader.fieldnames is not None else []
        self._validate_headers(fieldnames)

        # Step 3: Validate 'Source' and 'SYSTraderTranType' values
        self._validate_source_and_trader_type(reader)

        return file

    @staticmethod
    def _get_max_repeat(fieldnames: Sequence[str], section_prefix: str) -> int:
        """Identify the maximum number of repeats for a section."""
        max_repeat = 0
        for field in fieldnames:
            if field.startswith(section_prefix):
                try:
                    repeat_number = int(field.split("/")[-1])
                    max_repeat = max(max_repeat, repeat_number)
                except ValueError:
                    continue
        return max_repeat

    @staticmethod
    def _validate_file_type(file: UploadedFile) -> None:
        """Validate that the uploaded file is a CSV."""
        if not file.name.endswith(".csv"):
            msg = "File must be in CSV format."
            raise forms.ValidationError(msg)

    def _validate_headers(self, fieldnames: Sequence[str]) -> None:
        """Validate required and repeating columns in the headers."""
        missing_columns = [col for col in self.required_columns if col not in fieldnames]

        for section_name, column_list in self.repeating_columns.items():
            max_repeat = self._get_max_repeat(fieldnames, section_name)
            if max_repeat == 0:
                missing_columns.extend([f"{base_col}/1" for base_col in column_list])
            else:
                for repeat in range(1, max_repeat + 1):
                    missing_columns.extend([
                        f"{base_col}/{repeat}" for base_col in column_list if f"{base_col}/{repeat}" not in fieldnames
                    ])

        if missing_columns:
            msg = f"Missing required columns: {', '.join(missing_columns)}"
            raise forms.ValidationError(msg)

    @staticmethod
    def _validate_source_and_trader_type(data: Iterable[dict]) -> None:
        """Validate that 'Source' is always 80 and 'SYSTraderTranType' is always 4."""
        errors = []

        for index, row in enumerate(data, start=1):
            if row.get("Source") != "80":
                errors.append(f"Row {index}: 'Source' must be 80, but found {row.get('Source')}.")

            if row.get("SYSTraderTranType") != "4":
                errors.append(f"Row {index}: 'SYSTraderTranType' must be 4, but found {row.get('SYSTraderTranType')}.")

        if errors:
            raise forms.ValidationError(errors)
